package cn.com.qmth.examcloud.core.oe.task.service.job;

import cn.com.qmth.examcloud.commons.util.JsonMapper;
import cn.com.qmth.examcloud.core.oe.student.dao.ExamCaptureQueueRepo;
import cn.com.qmth.examcloud.support.CacheConstants;
import cn.com.qmth.examcloud.web.helpers.SequenceLockHelper;
import com.fasterxml.jackson.databind.JsonNode;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

/**
 * 人脸抓拍照片比对的执行次数复位归零任务
 */
@Component
public class FaceVerifyRestoreJobHandler {

    private static final Logger log = LoggerFactory.getLogger(FaceVerifyRestoreJobHandler.class);

    @Autowired
    private ExamCaptureQueueRepo examCaptureQueueRepo;

    public void run(int shardTotal, int shardIndex, String jobParam) throws Exception {
        long startTime = System.currentTimeMillis();
        int maxErrorNum = this.parseJobParam(jobParam);

        final String lockKey = CacheConstants.LOCK_FACE_COMPARE_RESET;
        try {
            SequenceLockHelper.getLockSimple(lockKey);

            // 补偿机制：将超过“最大错误执行次数”的记录重置为0，让这些记录可继续再重试执行
            int resetCount = examCaptureQueueRepo.resetExamCaptureQueueErrorNum(maxErrorNum);
            long cost = System.currentTimeMillis() - startTime;
            log.warn("人脸抓拍照片比对的执行次数复位归零任务_{}_{} 影响条数:{} maxErrorNum:{} cost:{}ms",
                    shardTotal, shardIndex, resetCount, maxErrorNum, cost);
        } catch (Exception e) {
            log.error(e.getMessage());
        } finally {
            SequenceLockHelper.releaseLockSimple(lockKey);
        }
    }

    private int parseJobParam(String jobParam) {
        // jobParam --> {"maxErrorNum":10}
        JsonNode jsonParams = new JsonMapper().getNode(jobParam);
        if (jsonParams == null) {
            return 1;
        }

        JsonNode maxErrorNum = jsonParams.get("maxErrorNum");
        if (maxErrorNum != null) {
            // 限制：最小值为1
            return Math.max(maxErrorNum.asInt(), 1);
        }

        return 1;
    }

}
